program plf;
uses
  Windows,
  SysUtils,
  Classes,
  Db,
  DbTables,
  ComUtil,
  ZLib in 'Zlib.pas',
  TextUtil in 'TextUtil.pas',
  U_PreLoad in 'U_PreLoad.pas';

{$R *.RES}

const
  ExeName:string='';

procedure Help(const Error: string);
begin
  if Error<>'' then Writeln('Error: ',Error);
  Writeln('Usage:');
  Writeln(ExeName,' archive.PLF -flags args [-flags args...]');
  Writeln('Flags are:');
  Writeln('  -l                          List contents of archive');
  Writeln('  -e ItemName TableName       Extract item from archive');
  Writeln('  -a ItemName [TableName]     Add item to archive');
  Writeln('  -d ItemName                 Delete item from archive');
  Writeln('  -n                          Create archive... (must be 2nd param!)');
  Writeln('  -p                          Pack archive... (must be 2nd and last param!)');
  Halt(ord(Error<>''));
end;

function GetParam(id: integer): string;
begin
  Result:=ParamStr(id);
  if (Result<>'') and (Result[1]='/') then Result[1]:='-';
end;

var
  ArcName: string;
  Archive: TDbPreLoadWriter=nil;
  NewArchive:Boolean;

procedure OpenArc;
begin
  if Archive=nil then begin
    if NewArchive then begin
      DeleteFile(ArcName);
      NewArchive:=False;
    end;
    Archive:=TDbPreLoadWriter.Create(ArcName);
  end;
end;

procedure GetFileName(var FileName: string; out is_db: Boolean);
begin
  if not FileExists(FileName) then begin
    FileName:=ChangeFileExt(FileName,'.DB');
    is_db:=True;
    if not FileExists(FileName) then begin
      is_db:=False;
      FileName:=ChangeFileExt(FileName,'.CSV');
      if not FileExists(FileName) then begin
        FileName:=ChangeFileExt(FileName,'.TXT');
        if not FileExists(FileName) then begin
          FileName:=ChangeFileExt(FileName,'.STX');
          if not FileExists(FileName) then begin
            Help('File not found:'+ChangeFileExt(FileName,'.???'));
          end;
        end;
      end;
    end;
  end else
    is_db:=SameText(ExtractFileExt(FileName),'.DB');
end;

procedure AddArc(const ItemName: string; FileName: string);
var is_db: Boolean;
    Stm: TStream;
    Tbl: TTable;
begin
  OpenArc;
  GetFileName(FileName,is_db);
  if is_db then begin
    Tbl:=TTable.Create(nil);
    try
      Tbl.DatabaseName:=ExtractFilePath(ExpandFileName(FileName));
      Tbl.TableName:=ExtractFileName(FileName);
      Tbl.Open;
      Archive.StoreTable(ItemName,Tbl);
    finally Tbl.Free; end;
    Writeln('Added item ',ItemName,' from table ',FileName);
  end else begin
    Stm:=TFileStream.Create(FileName,fmOpenRead or fmShareDenyNone);
    try
      Archive.StoreItem(ItemName,Stm,FileDateToDateTime(FileAge(FileName)));
    finally Stm.Free; end;
    Writeln('Added item '+ItemName,' from file ',FileName);
  end;
end;

procedure ExtractArc(const ItemName: string; FileName: string);
var is_db: Boolean;
    Stm,ArcStm: TStream;
    Tbl: TTable;
    Index,Size: integer;
    Ext,info: string;
begin
  OpenArc;
  //GetFileName(FileName,is_db);
  Ext:=ExtractFileExt(FileName);
  if (Ext='') and FileExists(FileName+'.DB') then begin
    Ext:='.DB';
    FileName:=FileName+'.DB';
  end;
  info:='';
  is_db:=SameText(Ext,'.db');
  if is_db then begin
    Tbl:=TTable.Create(nil);
    try
      Tbl.DatabaseName:=ExtractFilePath(ExpandFileName(FileName));
      Tbl.TableName:=ExtractFileName(FileName);
      Tbl.Exclusive:=True;
      Tbl.EmptyTable;
      Tbl.Open;
      Archive.LoadTable(ItemName,Tbl);
    finally Tbl.Free; end;
    Writeln('Extracted item ',ItemName,' into table ',FileName);
  end else begin
    Index:=Archive.IndexOfItem(ItemName);
    if (Index<0) or not Archive.OpenItem(Index,ArcStm,Size) then
      Help('Item '+ItemName+' not found in archive');
    try
      Stm:=TFileStream.Create(FileName,fmCreate);
      try
        ArcStm.Position:=0;
        if Size>0 then
          Stm.CopyFrom(ArcStm,Size)
        else
          info:=' (empty file)';
        FileSetDate(TFileStream(Stm).Handle,DateTimeToFileDate(Archive.ItemDate[Index]));
      finally Stm.Free; end;
    finally ArcStm.Free; end;
    Writeln('Extracted item ',ItemName,' into file ',FileName,info);
  end;
end;

procedure DeleteArc(const ItemName: string);
var Index: integer;
begin
  OpenArc;
  Index:=Archive.IndexOfItem(ItemName);
  if Index<0 then
    Writeln('Item ',ItemName,' does not exist!')
  else
    Archive.DeleteItem(Index);
end;

function Pad40(const s: string): string;
var l: integer;
begin
  Result:=s;
  l:=Length(Result);
  if l<40 then begin
    if (l and 7)<>0 then begin
      Result:=Result+#9;
      l:=(l+7) and (not 7);
    end;
    while l<40 do begin
      Result:=Result+#9;
      inc(l,8);
    end;
  end;
end;

procedure ListArchive;
var i: integer;
begin
  OpenArc;
  for i:=0 to Archive.ItemCount-1 do begin
    Writeln(Pad40(Archive.Items[i]),Archive.ItemSize[i],#9,DateToStr(Archive.ItemDate[i]));
  end;
end;

procedure PackArchive;
var NewArc: TPreLoadWriter;
    TempName,OldName: string;
    i: integer;
begin
  // Get unique temporary name:
  TempName:=ChangeFileExt(ArcName,'.~NEW');
  i:=0;
  while FileExists(TempName) do begin
    inc(i);
    TempName:=ChangeFileExt(TempName,'.~NEW_'+IntToStr(i));
  end;
  try
    OpenArc;
    NewArc:=TDbPreLoadWriter.Create(TempName);
    try
      NewArc.ImportArchive(Archive);
      FreeAndNil(Archive); // Close OLD!
    finally
      FreeAndNil(NewArc); // Close NEW!
    end;
    OldName:=ChangeFileExt(ArcName,'.~OLD');
    if not RenameFile(ArcName,OldName)
    or not RenameFile(TempName,ArcName) then begin
      // Restore old name:
      if not RenameFile(OldName,ArcName) then
        TempName:=''; // don't let it be deleted!
      raise Exception.Create('Cannot rename files!');
    end;
    DeleteFile(OldName);
  except
    DeleteFile(TempName);
    raise;
  end;
end;

const
  opAdd=1;
  opExtract=2;
  opDelete=3;

procedure DoIt;
var i,Cnt,op,nidx,nmcnt: integer;
    par: string;
    Names: array[0..1] of string;
    
  procedure Perform;
  begin
    case nidx of
      0: Help('Missing name');
      1: Names[1]:=Names[0];
    end;
    case Op of
      opAdd: begin
        AddArc(Names[0],Names[1]);
      end;
      opExtract: begin
        ExtractArc(Names[0],Names[1]);
      end;
      opDelete: begin
        DeleteArc(Names[0]);
      end;
    end;
  end;

begin
  Cnt:=ParamCount;
  ArcName:=GetParam(1);
  if SameText(ArcName,'-h') or SameText(ArcName,'/?') then Help('');
  if ParamCount<2 then Help('Not enough parameters!');
  i:=2;
  if ExtractFileExt(ArcName)='' then ArcName:=ArcName+'.PLF';
  if SameText(GetParam(i),'-n') then begin
    // Would create new archive...
    NewArchive:=True;
    inc(i);
  end else if SameText(GetParam(i),'-p') then begin
    if ParamCount>2 then Help('Too much parameters!');
    PackArchive;
    exit;
  end;
  if not FileExists(ArcName) then begin
    if not NewArchive then
      Help('Archive '+ArcName+' does not exist!');
  end;
  op:=0;
  nidx:=0; nmcnt:=2; // prevent warnings...
  while i<=Cnt do begin
    par:=GetParam(i);
    if (par='') then continue;
    if op=0 then begin
      nmcnt:=2;
      if False then
      else if SameText(par,'-a') then op:=opAdd
      else if SameText(par,'-e') then op:=opExtract
      else if SameText(par,'-d') then begin op:=opDelete; nmcnt:=1; end
      else if SameText(par,'-l') then ListArchive
      else Help('Unknown argument '+par);
      nidx:=0;
    end else begin
      if (par[1]='-') then begin
        Perform;
        op:=0;
        continue; // parse same flag again...
      end else begin
        Names[nidx]:=par;
        inc(nidx);
        if nidx>=nmcnt then begin
          Perform;
          op:=0;
        end;
      end;
    end;
    inc(i);
  end;
  if (op<>0) then
    Perform;
end;

var ECode:integer;

begin
  ECode:=0;
  ExeName:=ExtractFileName(ParamStr(0));
  try
    try
      DoIt;
    except on E:Exception do begin
      Writeln('Exception:');
      Writeln(E.Message);
      ECode:=1;
    end; end;
  finally FreeAndNil(Archive); end;
  Halt(ECode);
end.

